/* This file is part of Strigi Desktop Search
 *
 * Copyright (C) 2006 Jos van den Oever <jos@vandenoever.info>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */
#include <strigi/asyncsocket.h>
#include <sys/types.h>
#ifndef _WIN32_WCE
#include <sys/socket.h>
#include <sys/un.h>
#include <unistd.h>
#else
#include <winsock2.h>
#endif
#include <errno.h>
#include <string.h>
#include <stdio.h>

/* workaround for systems without MSG_NOSIGNAL
   on these systems the sending party will receive a SIGPIPE when the receiving
   party closes the connection */
#if defined(__APPLE__)
#define SOCKET_NOSIGNAL SO_NOSIGPIPE
#elif defined( MSG_NOSIGNAL )
#define SOCKET_NOSIGNAL MSG_NOSIGNAL
#else
#define SOCKET_NOSIGNAL 0
#endif

AsyncSocket::AsyncSocket() {
    status = Idle;
#ifdef _WIN32
    INT iRet = 0;
    WSADATA wsaData;
    iRet = WSAStartup(MAKEWORD(2,2), &wsaData);
    if (iRet != 0){
        perror("WSAStartup failed ");
    }
#endif
}
AsyncSocket::~AsyncSocket() {
    close();
#ifdef _WIN32
        WSACleanup();
#endif
}
void
AsyncSocket::open() {
#ifndef _WIN32
    struct sockaddr_un  serv_addr;

    // create a socket
    socket = ::socket(AF_UNIX, SOCK_STREAM, 0);
    if (socket < 0) {
        error = "Could not create socket: ";
        error += strerror(errno);
        return;
    }

    // set the address
    memset(&serv_addr, 0, sizeof(serv_addr));
    serv_addr.sun_family = AF_UNIX;
    size_t len = socketpath.length();
    len = (len > sizeof(serv_addr.sun_path)) ?sizeof(serv_addr.sun_path) :len;
    strncpy(serv_addr.sun_path, socketpath.c_str(), len);
    serv_addr.sun_path[len] = '\0';

    // connect to the server
    int r = connect(socket, (struct sockaddr *)&serv_addr, sizeof(serv_addr));
    if (r < 0) {
        error = "Could not connect to server: ";
        error += strerror(errno);
        ::close(socket);
        socket = -1;
    }
#else
    sockaddr_in serv_addr;

    // create a socket
    socket = (int)::socket( AF_INET, SOCK_STREAM, IPPROTO_TCP);
    if (socket == INVALID_SOCKET) {
        error = "Could not create socket: ";
        error += strerror(errno);
        return;
    }

    // set the address
    serv_addr.sin_family = AF_INET;
    serv_addr.sin_addr.s_addr = htonl (INADDR_LOOPBACK);
    serv_addr.sin_port = htons( 5555 );

    // connect to the server
    int r = connect((SOCKET)socket, (SOCKADDR*) &serv_addr, sizeof(serv_addr));
    if (r != 0) {
        error = "Could not connect to server: ";
        error += strerror(errno);
        ::closesocket((SOCKET)socket);
        socket = -1;
    }
#endif
}
void
AsyncSocket::close() {
    if (socket >= 0) {
#ifndef _WIN32
        ::close(socket);
#else
        closesocket((SOCKET)socket);
#endif
    }
}
bool
AsyncSocket::sendRequest(const std::string& req) {
    if (status == Reading || status == Writing) {
        close();
    }
    open();
    if (socket < 0) {
        status = Error;
        return false;
    }
    error.clear();
    request = req;
    response.clear();
    writepos = 0;
    status = Writing;
    return true;
}
bool
AsyncSocket::statusChanged() {
    if (status == Error || status == Idle) return false;
    if (status == Writing) {
        write();
    }
    if (status == Reading) {
        read();
    }
    return status == Error || status == Idle;
}
void
AsyncSocket::write() {
#ifndef _WIN32
    ssize_t r;
#else
    int r;
#endif
#ifndef _WIN32
    r = send(socket, request.c_str()+writepos,
#else
    r = send((SOCKET)socket, request.c_str()+writepos,
#endif
        request.length()-writepos, SOCKET_NOSIGNAL);
    if (r != -1) {
        writepos += (unsigned)r;
        if (writepos == request.length()) {
            status = Reading;
        }
    } else {
        switch (errno) {
        case EAGAIN:
            break;
        default:
            status = Error;
            error = strerror(errno);
            break;
        }
    }
}
void
AsyncSocket::read() {
    char c;
    while (true) {
#ifndef _WIN32
        ssize_t r;
#else
        int r;
#endif
#ifndef _WIN32_WCE
        r = recv(socket, &c, 1, MSG_DONTWAIT|SOCKET_NOSIGNAL);
#else
        r = recv(socket, &c, 1, 0);
#endif
        switch (r) {
        case 0:
            close();
            status = Idle;
            return;
        case 1:
            response += c;
            break;
        default:
#ifndef _WIN32_WCE
            if (errno != EAGAIN) {
                printf("error %i: %s\n", errno, strerror(errno));
#else
            if (r ==SOCKET_ERROR) {
                printf("error receiving\n");
#endif
                status = Error;
                error = strerror(EAGAIN);
                close();
            }
            return;
        }
    }
}
